@@ -10,9 +10,14 @@ use std::mem;
 use std::path::Path;
 use std::sync::Arc;
 use wasmparser::{Parser, ValidPayload, Validator};
-use wasmtime_environ::{ModuleEnvironment, ModuleIndex, PrimaryMap};
-use wasmtime_jit::{CompiledModule, CompiledModuleInfo, MmapVec, TypeTables};
-use wasmtime_runtime::ModuleMemFds;
+use wasmtime_environ::{
+    DefinedFuncIndex, DefinedMemoryIndex, FunctionInfo, ModuleEnvironment, ModuleIndex, PrimaryMap,
+    SignatureIndex,
+};
+use wasmtime_jit::{CompiledModule, CompiledModuleInfo, TypeTables};
+use wasmtime_runtime::{
+    CompiledModuleId, MemoryImage, MmapVec, ModuleMemoryImages, VMSharedSignatureIndex,
+};
 
 mod registry;
 mod serialization;
@@ -109,11 +114,10 @@ struct ModuleInner {
     types: Arc<TypeTables>,
     /// Registered shared signature for the module.
     signatures: Arc<SignatureCollection>,
-    /// A set of memfd images for memories, if any. Note that module
-    /// instantiation (hence the need for lazy init) may happen for the
-    /// same module concurrently in multiple Stores, so we use a
-    /// OnceCell.
-    memfds: OnceCell<Option<Arc<ModuleMemFds>>>,
+    /// A set of initialization images for memories, if any. Note that module
+    /// instantiation (hence the need for lazy init) may happen for the same
+    /// module concurrently in multiple Stores, so we use a OnceCell.
+    memory_images: OnceCell<Option<ModuleMemoryImages>>,
 }
 
 impl Module {
@@ -421,6 +425,21 @@ impl Module {
                 translation.try_paged_init();
             }
 
+            // If configured attempt to use static memory initialization which
+            // can either at runtime be implemented as a single memcpy to
+            // initialize memory or otherwise enabling virtual-memory-tricks
+            // such as mmap'ing from a file to get copy-on-write.
+            if engine.config().memory_init_cow {
+                let align = engine.compiler().page_size_align();
+                let max_always_allowed = engine.config().memory_guaranteed_dense_image_size;
+                translation.try_static_init(align, max_always_allowed);
+            }
+
+            // Attempt to convert table initializer segments to
+            // FuncTable representation where possible, to enable
+            // table lazy init.
+            translation.try_func_table_init();
+
             let (mmap, info) =
                 wasmtime_jit::finish_compile(translation, obj, funcs, trampolines, tunables)?;
             Ok((mmap, Some(info)))
@@ -487,14 +506,26 @@ impl Module {
     /// Same as [`deserialize`], except that the contents of `path` are read to
     /// deserialize into a [`Module`].
     ///
-    /// For more information see the documentation of the [`deserialize`]
-    /// method for why this function is `unsafe`.
-    ///
     /// This method is provided because it can be faster than [`deserialize`]
     /// since the data doesn't need to be copied around, but rather the module
     /// can be used directly from an mmap'd view of the file provided.
     ///
     /// [`deserialize`]: Module::deserialize
+    ///
+    /// # Unsafety
+    ///
+    /// All of the reasons that [`deserialize`] is `unsafe` applies to this
+    /// function as well. Arbitrary data loaded from a file may trick Wasmtime
+    /// into arbitrary code execution since the contents of the file are not
+    /// validated to be a valid precompiled module.
+    ///
+    /// Additionally though this function is also `unsafe` because the file
+    /// referenced must remain unchanged and a valid precompiled module for the
+    /// entire lifetime of the [`Module`] returned. Any changes to the file on
+    /// disk may change future instantiations of the module to be incorrect.
+    /// This is because the file is mapped into memory and lazily loaded pages
+    /// reflect the current state of the file, not necessarily the origianl
+    /// state of the file.
     pub unsafe fn deserialize_file(engine: &Engine, path: impl AsRef<Path>) -> Result<Module> {
         let module = SerializedModule::from_file(path.as_ref(), &engine.config().module_version)?;
         module.into_module(engine)
@@ -507,8 +538,10 @@ impl Module {
         types: Arc<TypeTables>,
         module_upvars: &[serialization::SerializedModuleUpvar],
     ) -> Result<Self> {
-        // Validate the module can be used with the current allocator
-        engine.allocator().validate(modules[main_module].module())?;
+        // Validate all modules can be used with the current allocator
+        for module in modules.iter() {
+            engine.allocator().validate(module.module())?;
+        }
 
         let signatures = Arc::new(SignatureCollection::new_for_module(
             engine.signatures(),
@@ -533,7 +566,7 @@ impl Module {
                     &signatures,
                 )
             })
-            .collect::<Result<Vec<_>>>()?;
+            .collect();
 
         return Ok(Self {
             inner: Arc::new(ModuleInner {
@@ -543,7 +576,7 @@ impl Module {
                 artifact_upvars: modules,
                 module_upvars,
                 signatures,
-                memfds: OnceCell::new(),
+                memory_images: OnceCell::new(),
             }),
         });
 
@@ -555,14 +588,14 @@ impl Module {
             artifact_upvars: &[usize],
             module_upvars: &[serialization::SerializedModuleUpvar],
             signatures: &Arc<SignatureCollection>,
-        ) -> Result<Module> {
+        ) -> Module {
             let module = artifacts[module_index].clone();
-            Ok(Module {
+            Module {
                 inner: Arc::new(ModuleInner {
                     engine: engine.clone(),
                     types: types.clone(),
                     module,
-                    memfds: OnceCell::new(),
+                    memory_images: OnceCell::new(),
                     artifact_upvars: artifact_upvars
                         .iter()
                         .map(|i| artifacts[*i].clone())
@@ -580,10 +613,10 @@ impl Module {
                                 signatures,
                             )
                         })
-                        .collect::<Result<Vec<_>>>()?,
+                        .collect(),
                     signatures: signatures.clone(),
                 }),
-            })
+            }
         }
     }
 
@@ -688,7 +721,7 @@ impl Module {
                 types: self.inner.types.clone(),
                 engine: self.inner.engine.clone(),
                 module,
-                memfds: OnceCell::new(),
+                memory_images: OnceCell::new(),
                 artifact_upvars: artifact_upvars
                     .iter()
                     .map(|i| self.inner.artifact_upvars[*i].clone())
@@ -723,16 +756,6 @@ impl Module {
         &self.inner.signatures
     }
 
-    pub(crate) fn memfds(&self) -> Result<Option<&Arc<ModuleMemFds>>> {
-        Ok(self
-            .inner
-            .memfds
-            .get_or_try_init(|| {
-                ModuleMemFds::new(self.inner.module.module(), self.inner.module.wasm_data())
-            })?
-            .as_ref())
-    }
-
     /// Looks up the module upvar value at the `index` specified.
     ///
     /// Note that this panics if `index` is out of bounds since this should
@@ -950,6 +973,14 @@ impl Module {
     pub fn engine(&self) -> &Engine {
         &self.inner.engine
     }
+
+    /// Returns the `ModuleInner` cast as `ModuleRuntimeInfo` for use
+    /// by the runtime.
+    pub(crate) fn runtime_info(&self) -> Arc<dyn wasmtime_runtime::ModuleRuntimeInfo> {
+        // N.B.: this needs to return a clone because we cannot
+        // statically cast the &Arc<ModuleInner> to &Arc<dyn Trait...>.
+        self.inner.clone()
+    }
 }
 
 fn _assert_send_sync() {
@@ -984,3 +1015,146 @@ impl std::hash::Hash for HashedEngineCompileEnv<'_> {
         env!("CARGO_PKG_VERSION").hash(hasher);
     }
 }
+
+impl wasmtime_runtime::ModuleRuntimeInfo for ModuleInner {
+    fn module(&self) -> &Arc<wasmtime_environ::Module> {
+        self.module.module()
+    }
+
+    fn signature(&self, index: SignatureIndex) -> VMSharedSignatureIndex {
+        self.signatures.as_module_map()[index]
+    }
+
+    fn image_base(&self) -> usize {
+        self.module.code().as_ptr() as usize
+    }
+
+    fn function_info(&self, index: DefinedFuncIndex) -> &FunctionInfo {
+        self.module.func_info(index)
+    }
+
+    fn memory_image(&self, memory: DefinedMemoryIndex) -> Result<Option<&Arc<MemoryImage>>> {
+        if !self.engine.config().memory_init_cow {
+            return Ok(None);
+        }
+
+        let images = self.memory_images.get_or_try_init(|| {
+            ModuleMemoryImages::new(
+                self.module.module(),
+                self.module.wasm_data(),
+                Some(self.module.mmap()),
+            )
+        })?;
+        Ok(images
+            .as_ref()
+            .and_then(|images| images.get_memory_image(memory)))
+    }
+
+    fn unique_id(&self) -> Option<CompiledModuleId> {
+        Some(self.module.unique_id())
+    }
+
+    fn wasm_data(&self) -> &[u8] {
+        self.module.wasm_data()
+    }
+
+    fn signature_ids(&self) -> &[VMSharedSignatureIndex] {
+        self.signatures.as_module_map().values().as_slice()
+    }
+}
+
+/// A barebones implementation of ModuleRuntimeInfo that is useful for
+/// cases where a purpose-built environ::Module is used and a full
+/// CompiledModule does not exist (for example, for tests or for the
+/// default-callee instance).
+pub(crate) struct BareModuleInfo {
+    module: Arc<wasmtime_environ::Module>,
+    image_base: usize,
+    one_signature: Option<(SignatureIndex, VMSharedSignatureIndex)>,
+    function_info: PrimaryMap<DefinedFuncIndex, FunctionInfo>,
+}
+
+impl BareModuleInfo {
+    pub(crate) fn empty(module: Arc<wasmtime_environ::Module>) -> Self {
+        BareModuleInfo {
+            module,
+            image_base: 0,
+            one_signature: None,
+            function_info: PrimaryMap::default(),
+        }
+    }
+
+    pub(crate) fn maybe_imported_func(
+        module: Arc<wasmtime_environ::Module>,
+        one_signature: Option<(SignatureIndex, VMSharedSignatureIndex)>,
+    ) -> Self {
+        BareModuleInfo {
+            module,
+            image_base: 0,
+            one_signature,
+            function_info: PrimaryMap::default(),
+        }
+    }
+
+    pub(crate) fn one_func(
+        module: Arc<wasmtime_environ::Module>,
+        image_base: usize,
+        info: FunctionInfo,
+        signature_id: SignatureIndex,
+        signature: VMSharedSignatureIndex,
+    ) -> Self {
+        let mut function_info = PrimaryMap::with_capacity(1);
+        function_info.push(info);
+        BareModuleInfo {
+            module,
+            image_base,
+            function_info,
+            one_signature: Some((signature_id, signature)),
+        }
+    }
+
+    pub(crate) fn into_traitobj(self) -> Arc<dyn wasmtime_runtime::ModuleRuntimeInfo> {
+        Arc::new(self)
+    }
+}
+
+impl wasmtime_runtime::ModuleRuntimeInfo for BareModuleInfo {
+    fn module(&self) -> &Arc<wasmtime_environ::Module> {
+        &self.module
+    }
+
+    fn signature(&self, index: SignatureIndex) -> VMSharedSignatureIndex {
+        let (signature_id, signature) = self
+            .one_signature
+            .expect("Signature for one function should be present if queried");
+        assert_eq!(index, signature_id);
+        signature
+    }
+
+    fn image_base(&self) -> usize {
+        self.image_base
+    }
+
+    fn function_info(&self, index: DefinedFuncIndex) -> &FunctionInfo {
+        &self.function_info[index]
+    }
+
+    fn memory_image(&self, _memory: DefinedMemoryIndex) -> Result<Option<&Arc<MemoryImage>>> {
+        Ok(None)
+    }
+
+    fn unique_id(&self) -> Option<CompiledModuleId> {
+        None
+    }
+
+    fn wasm_data(&self) -> &[u8] {
+        &[]
+    }
+
+    fn signature_ids(&self) -> &[VMSharedSignatureIndex] {
+        match &self.one_signature {
+            Some((_, id)) => std::slice::from_ref(id),
+            None => &[],
+        }
+    }
+}
